import json
import pathlib
import sys

import click

from payrange import _scan, _characteristics

cli = click.Group()


def _print_json(obj):
    print(json.dumps(obj, indent=2, sort_keys=True))


@cli.command()
@click.option('-m', '--mac-prefix', default=None, type=str, help='Only include devices with MAC address prefix.')
@click.option('-p', '--name-prefix', default=None, type=str, help='Only include devices with name prefix.')
@click.option('-n', '--count', default=None, type=int, help='Stop after finding N devices.')
@click.option('-t', '--duration', default=None, type=float, help='Stop after T seconds.')
def scan(mac_prefix, name_prefix, count, duration):
    """Scan for BLE devices."""
    results = _scan.scan(mac_prefix, name_prefix, count, duration)
    _print_json(results)


@cli.command()
@click.option('-m', '--mac-prefix', default=None, type=str, help='Only include devices with MAC address prefix.')
@click.option('-p', '--name-prefix', default=None, type=str, help='Only include devices with name prefix.')
@click.option('-n', '--count', default=None, type=int, help='Stop after finding N devices.')
@click.option('-t', '--duration', default=None, type=float, help='Stop after T seconds.')
def rssi(mac_prefix, name_prefix, count, duration):
    """Collect RSSI statistics for BLE devices."""
    results = _scan.collect_rssi_stats(mac_prefix, name_prefix, count, duration)
    _print_json(results)


@cli.command()
@click.option('-m', '--mac-prefix', default=None, type=str, help='Only include devices with MAC address prefix.')
@click.option('-p', '--name-prefix', default=None, type=str, help='Only include devices with name prefix.')
@click.option('-n', '--count', default=None, type=int, help='Stop after finding N devices.')
@click.option('-t', '--duration', default=None, type=float, help='Stop after T seconds.')
def mfgdata(mac_prefix, name_prefix, count, duration):
    """Capture manufacturer specific data transmitted in advertising packets."""
    results = _scan.capture_manufacturer_data(mac_prefix, name_prefix, count, duration)
    _print_json(results)


@cli.command()
@click.argument('address', type=str)
def findchars(address: str):
    """Discover BLE characteristics available on a given device."""
    results = _characteristics.find_characteristics(address)
    _print_json(results)


@cli.command()
@click.argument('address', type=str, nargs=1)
@click.argument('characteristics', type=str, nargs=-1)
def readchars(address: str, characteristics: tuple[str]):
    """Read characteristic values from one or many devices.

    Either argument may be replaced with a file containing the output of `scan` or `findchars`."""
    addr_path = pathlib.Path(address)
    if addr_path.exists():
        print('Loading addresses from file...', file=sys.stderr)
        with addr_path.open('r') as f:
            contents = json.load(f)
        addresses = list(contents.keys())
    else:
        addresses = [address]

    if len(characteristics) == 1:
        chars_path = pathlib.Path(characteristics[0])
        if chars_path.exists():
            print('Loading characteristics from file...', file=sys.stderr)
            with chars_path.open('r') as f:
                contents = json.load(f)
            characteristics = [char for char, props in contents.items()
                               if 'READ' in set(props)]

    results = {
        address: _characteristics.read_characteristics(address, characteristics)
        for address in addresses
    }
    _print_json(results)


if __name__ == '__main__':
    cli()
