import binascii
import sys

from bluepy.btle import Characteristic, Peripheral, BTLEException, UUID

from payrange._common import execute_with_retry, CharacteristicMatcher

__all__ = [
    'find_characteristics',
    'read_characteristics'
]


def find_characteristics(address: str) -> dict[str, list[str]]:
    """Find all available characteristics at an address.

    :param address: A MAC address to connect to
    :returns: A dictionary mapping each characteristic name to a list of properties.
    """

    address = address.strip().casefold()
    peripheral = Peripheral()
    chars = execute_with_retry(peripheral, address, lambda: peripheral.getCharacteristics())
    try:
        peripheral.disconnect()
    except BTLEException:
        pass

    return {char.uuid.getCommonName():
            [name for prop, name in Characteristic.propNames.items()
             if char.properties & prop]
            for char in chars}


def read_characteristics(address: str, char_names: list[str]) -> dict[str, str]:
    """Read for a list of characteristic names.

    :param address: A MAC address to connect to
    :param char_names: A list of characteristic names to read, which may be either
        common names or UUIDs.
    :returns: A dictionary mapping each characteristic to its value.
    """

    if len(char_names) == 0:
        return {}

    address = address.strip().casefold()
    peripheral = Peripheral()

    print('Looking up characteristics...', file=sys.stderr)
    chars = execute_with_retry(peripheral, address, lambda: peripheral.getCharacteristics())

    common_name_to_char = {char.uuid.getCommonName().replace('-', '').replace('_', '').casefold(): char
                           for char in chars}

    match_uuid = CharacteristicMatcher(chars)

    arg_to_char = {}
    for arg in char_names:
        char_name = arg.strip().replace('-', '').replace('_', '').casefold()

        # char_name might be a Common Name or non-standard UUID
        if char_name in common_name_to_char:
            arg_to_char[arg] = common_name_to_char[char_name]
            continue

        # char_name might also be a standard UUID
        maybe_char = match_uuid(char_name)
        if maybe_char is not None:
            arg_to_char[arg] = maybe_char

        print('Could not parse characteristic:', arg, file=sys.stderr)

    results = {}
    for arg, char in arg_to_char.items():
        if not char.supportsRead():
            print('Characteristic', char.uuid.getCommonName(), 'not readable.', file=sys.stderr)
            continue

        print('Reading characteristic', arg, file=sys.stderr)
        value = execute_with_retry(peripheral, address, lambda : char.read())
        print('\tValue:', value, file=sys.stderr)

        try:
            str_value = value.decode('ascii')
            if not str_value.isprintable():
                str_value = '0x' + binascii.b2a_hex(value).decode('ascii').upper()
        except UnicodeDecodeError:
            str_value = '0x' + binascii.b2a_hex(value).decode('ascii').upper()
        results[arg] = str_value

    return results
