import sys
from typing import Callable, TypeVar, Optional

import binascii

from bluepy.btle import BTLEInternalError, Peripheral, BTLEException, \
    BTLEDisconnectError, Characteristic, UUID, ScanEntry

T = TypeVar('T')

RECONNECT_ATTEMPTS = 25


def execute_with_retry(peripheral: Peripheral, address: str, command: Callable[[], T]) -> T:
    for attempt in range(RECONNECT_ATTEMPTS):
        try:
            return command()
        except BTLEInternalError as exc:
            print(f'Attempting to connect... ({attempt + 1}/{RECONNECT_ATTEMPTS})', file=sys.stderr)
            try:
                peripheral.connect(address)
            except BTLEException:
                pass
        except BTLEDisconnectError as exc:
            print(f'Device disconnected. Retrying... ({attempt + 1}/{RECONNECT_ATTEMPTS})', file=sys.stderr)
            try:
                peripheral.connect(address)
            except BTLEException:
                pass
    else:
        raise RuntimeError('Failed to connect (retries exhausted)')


class ScanEntryFilter(object):

    def __init__(self, mac_prefix: Optional[str] = None, name_prefix: Optional[str] = None):
        if mac_prefix is not None:
            mac_prefix = mac_prefix.strip().casefold().replace(':', '')

        if name_prefix is not None:
            name_prefix = name_prefix.strip().casefold()

        self._mac_prefix = mac_prefix
        self._name_prefix = name_prefix

    def __call__(self, scan_entry: ScanEntry) -> bool:
        mac_address = scan_entry.addr
        device_name: str | None = \
            scan_entry.getValue(ScanEntry.COMPLETE_LOCAL_NAME) or \
            scan_entry.getValue(ScanEntry.SHORT_LOCAL_NAME)

        mac_address = mac_address.casefold().replace(':', '')
        if self._mac_prefix is not None and not mac_address.startswith(self._mac_prefix):
            return False

        if self._name_prefix is not None:
            if device_name is None:
                return False

            return device_name.casefold().strip().startswith(self._name_prefix)

        return True


class CharacteristicMatcher(object):

    def __init__(self, characteristics: list[Characteristic]):
        self._chars = characteristics

    def __call__(self, target: str) -> Optional[Characteristic]:
        try:
            uuid = UUID(target)
        except binascii.Error:
            return None

        for char in self._chars:
            if uuid == char.uuid:
                return char

        return None

