from abc import abstractmethod, ABC
import collections
import statistics
import sys
import time
from typing import TypeVar, Generic

from bluepy.btle import Scanner, DefaultDelegate, ScanEntry, BTLEManagementError

from payrange._common import ScanEntryFilter
from payrange._rle import RunLengthEncoder

__all__ = [
    'scan'
]


T = TypeVar('T')


class ScanDelegateBase(DefaultDelegate, ABC, Generic[T]):

    def __init__(self, count: int | None = None, duration: float | None = None) -> None:
        super().__init__()
        self._duration = duration
        self._max_count = count
        self._start_time = None

    @property
    def should_stop(self) -> bool:
        if self._max_count is not None and self.count >= self._max_count:
            return True

        if self._duration is not None and time.monotonic() - self._start_time >= self._duration:
            return True

        return False

    @property
    @abstractmethod
    def results(self) -> T:
        ...

    @property
    @abstractmethod
    def count(self) -> int:
        ...

    def start(self) -> None:
        self._start_time = time.monotonic()

    # noinspection PyPep8Naming
    @abstractmethod
    def handleDiscovery(self, scan_entry: ScanEntry, is_new_device: bool, is_new_data: bool) -> None:
        ...


class DeviceScanDelegate(ScanDelegateBase[dict[str, str]]):

    def __init__(self,  mac_prefix: str | None = None, name_prefix: str | None = None,
                 count: int | None = None, duration: float | None = None) -> None:
        super().__init__(count, duration)
        self._filter_predicate = ScanEntryFilter(mac_prefix, name_prefix)
        self._results = {}

    def handleDiscovery(self, scan_entry: ScanEntry, is_new_device: bool, is_new_data: bool) -> None:
        if not is_new_device or not is_new_data:
            return
        
        if not self._filter_predicate(scan_entry):
            return
        
        mac = scan_entry.addr
        name = scan_entry.getValue(ScanEntry.COMPLETE_LOCAL_NAME) or scan_entry.getValue(ScanEntry.SHORT_LOCAL_NAME)

        if mac not in self._results:
            print('Found', mac, name or '(unknown name)', file=sys.stderr)

        self._results[mac] = name or self._results.get(mac, None)

    @property
    def count(self):
        return len(self._results)

    @property
    def results(self) -> T:
        return self._results


class RSSIScanDelegate(ScanDelegateBase[dict[str, list[float]]]):

    def __init__(self, mac_prefix: str | None = None, name_prefix: str | None = None,
                 count: int | None = None, duration: float | None = None) -> None:
        super().__init__(count, duration)
        self._filter_predicate = ScanEntryFilter(mac_prefix, name_prefix)
        self._results = collections.defaultdict(list)

    def handleDiscovery(self, scan_entry: ScanEntry, is_new_device: bool, is_new_data: bool) -> None:
        if not self._filter_predicate(scan_entry):
            return

        mac = scan_entry.addr
        self.results[mac].append(scan_entry.rssi)
        print(f'{mac} -> {scan_entry.rssi} (x{len(self.results[mac])})', file=sys.stderr)

    @property
    def count(self):
        return min((len(values) for values in self.results.values()), default=0)

    @property
    def results(self) -> T:
        return self._results


class MfgDataScanDelegate(ScanDelegateBase[dict[str, RunLengthEncoder[bytes]]]):

    def __init__(self, mac_prefix: str | None = None, name_prefix: str | None = None,
                 max_count: int | None = None, duration: float | None = None) -> None:
        super().__init__(max_count, duration)
        self._filter_predicate = ScanEntryFilter(mac_prefix, name_prefix)
        self._results = collections.defaultdict(RunLengthEncoder)

    def handleDiscovery(self, scan_entry: ScanEntry, is_new_device: bool, is_new_data: bool) -> None:
        if not self._filter_predicate(scan_entry):
            return

        mac = scan_entry.addr
        mfg = scan_entry.getValueText(ScanEntry.MANUFACTURER)
        if mfg is None:
            return

        mfg = '0x' + mfg.upper()
        self._results[mac].append(mfg)
        print(f'{mac} -> {mfg} (x{len(self._results[mac])})', file=sys.stderr)

    @property
    def count(self) -> int:
        return min((len(values) for values in self._results.values()), default=0)

    @property
    def results(self) -> T:
        return self._results


def _perform_scan_with_delegate(delegate: ScanDelegateBase[T]) -> T:

    scanner = Scanner().withDelegate(delegate)

    # noinspection PyBroadException
    try:
        scanner.stop()
    except Exception:
        pass

    try:
        print('Starting scan...', file=sys.stderr)
        delegate.start()
        scanner.start()
        while True:
            scanner.process(0.01)

            if delegate.should_stop:
                break

    except KeyboardInterrupt:
        pass

    except BTLEManagementError as exc:
        if exc.estat == 20:
            raise RuntimeError('Re-run as root!')
        else:
            raise

    finally:
        # noinspection PyBroadException
        try:
            scanner.stop()
        except Exception:
            pass

    print('Stopped scan.', file=sys.stderr)
    return delegate.results


def scan(mac_prefix: str = None, name_prefix: str = None, count: int = None, duration: float = None) -> dict[str, str]:
    return _perform_scan_with_delegate(DeviceScanDelegate(mac_prefix, name_prefix, count, duration))


def collect_rssi_stats(mac_prefix: str = None, name_prefix: str = None, count: int = None, duration: float = None) -> \
        dict[str, dict[str, float]]:

    results = _perform_scan_with_delegate(RSSIScanDelegate(mac_prefix, name_prefix, count, duration))

    return {
        mac: {
            'min': min(vals),
            'max': max(vals),
            'mean': statistics.mean(vals),
            'median': statistics.median(vals),
            'variance': statistics.variance(vals) if len(vals) > 1 else 0.0,
            'n': len(vals)
        }
        for mac, vals in results.items()
    }


def capture_manufacturer_data(mac_prefix: str = None, name_prefix: str = None, count: int = None,
                              duration: float = None) -> dict[str, list]:
    results = _perform_scan_with_delegate(MfgDataScanDelegate(mac_prefix, name_prefix, count, duration))

    return {
        addr: list(values.values)
        for addr, values in results.items()
    }
